#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Date    : 2018-03-13 10:34:43
# @Author  : Zhi Liu (zhiliu.mind@gmail.com)
# @Link    : http://iridescent.ink
# @Version : $1.0$
from __future__ import division, print_function, absolute_import

# DEFINE GLOBAL CONSTS
C = 299792458.0  # 2.99792458e8 in m/s
# PI = 3.14159265358979323846
PI = 3.141592653589793238462643383279502884197169399375105820974944592307
EPS = 1.0e-32


Ge = 398603e9  # the constant of earth gravitation
Rea = 6367.856e3  # the avarage radius of earth
Ree = 6378.137e3  # the equatorial radius of earth
Rep = 6356.752e3  # the polar radius of earth
